"""NLP Result class.

Should be a return type of all NL processors.
"""
from typing import Any, Callable, Tuple

from voiceassistant.interfaces import InterfaceType
from voiceassistant.utils.datastruct import DottedDict


class NlpResult:
    """NLP Result class."""

    def __init__(
        self,
        skill_name: str,
        skill_func: Callable,
        entities: DottedDict,
        is_complete: bool,
    ):
        """Init."""
        self.skill_name = skill_name
        self.skill_func = skill_func
        self.entities = entities
        self.is_complete = is_complete

    def __str__(self) -> str:
        """Get string respesentation of NLP result."""
        return "\n".join(
            (
                "_" * 20,
                "NLP Result:",
                f"skill:    {self.skill_name}",
                f"entities: {self.entities}",
                f"complete: {self.is_complete}",
                "_" * 20,
            )
        )

    def _key(self) -> Tuple:
        """Get NLP result key."""
        return (
            tuple((k, v) for k, v in self.entities.items()),
            self.skill_func,
            self.is_complete,
        )

    def __hash__(self) -> int:
        """Get hash of this object."""
        return hash(self._key())

    def __eq__(self, other: Any) -> bool:
        """Check equality with another NlpResult instance."""
        return self._key() == other._key()  # type: ignore

    def execute_skill(self, interface: InterfaceType) -> None:
        """Execute skill function from NLP result."""
        print(self)
        self.skill_func(entities=self.entities, interface=interface)



"""Add-On functions for triggerword detection event."""

import os
import random
import subprocess
from functools import lru_cache

from voiceassistant.config import Config
from voiceassistant.const import DATA_DIR, DEFAULT_CONFIG_DIR
from voiceassistant.exceptions import ConfigValidationError


def _react_by_random_phrase(vass) -> None:  # type: ignore
    """React to trigger word by a random reply phrase."""
    vass.speech.output(
        text=random.choice(Config.triggerword.replies), cache=True,
    )


def _react_by_sound(vass) -> None:  # type: ignore
    """React to trigger detection word."""
    subprocess.Popen(["mpg123", _get_soundfile_path(Config.triggerword.sound)])


@lru_cache
def _get_soundfile_path(name: str) -> str:
    """Get triggerword reaction sound path."""
    path_preset = f"{DATA_DIR}/audio/{name}.mp3"
    path_custom = f"{DEFAULT_CONFIG_DIR}/audio/{name}.mp3"

    if os.path.isfile(path_preset):
        return path_preset
    elif os.path.isfile(path_custom):
        return path_custom
    else:
        raise ConfigValidationError(f"sound file '{name}' doesn't exist")


def _do_not_react(vass) -> None:  # type: ignore
    """Do nothing if reaction is not set in Config."""
    print("Keyword detected")


if "sound" in Config.triggerword:
    react_to_keyword = _react_by_sound
elif "replies" in Config.triggerword:
    react_to_keyword = _react_by_random_phrase
else:
    react_to_keyword = _do_not_react



"""Host add-ons for different components."""

from typing import Any, Callable, Optional

# flake8: noqa
# from . import speech
# from . import keyword


def call_at(
    start: Optional[Callable] = None, end: Optional[Callable] = None
) -> Callable:
    """Wrap func to call specified function at the start and end."""

    def wrapper(func: Callable) -> Callable:
        def inner(*args: Any, **kwargs: Any) -> Any:
            if start is not None:
                start()

            result = func(*args, **kwargs)

            if end is not None:
                end()

            return result

        return inner

    return wrapper



"""API app factory."""

from flask import Flask, Response, jsonify, request

from voiceassistant.config import Config
from voiceassistant.exceptions import ConfigValidationError
from voiceassistant.nlp.regex import REGEX_SKILLS
from voiceassistant.utils.datastruct import DottedDict

from .auth import authorized


def api_factory(vass, app: Flask) -> Flask:  # type: ignore
    """Get REST API app."""
    name = "api"

    @app.route(f"/{name}/say", methods=["POST"])
    @authorized
    def say() -> Response:
        """Pronounce text.

        Sample payload:
        {"text": "Hello, World"}
        """
        payload = request.get_json() or {}
        try:
            vass.speech.output(payload["text"])
            return Response(status=200)
        except (KeyError, TypeError):
            return Response("Payload must have 'text' key", status=406)

    @app.route(f"/{name}/skills", methods=["GET"])
    @authorized
    def get_skills() -> Response:
        """Get an array of all available skill names."""
        skillnames = list(REGEX_SKILLS.keys())
        return jsonify(skillnames)

    @app.route(f"/{name}/skills/<skill_name>", methods=["GET", "POST"])
    @authorized
    def execute_skill(skill_name: str) -> Response:
        """Execure specified skill.

        Sample payload (Optional):
        {"entities": {"location": "Moscow"}}
        """
        payload = request.get_json() or {}
        try:
            entities = payload.get("entities", {})
            skill_func = REGEX_SKILLS[skill_name].func
            skill_func(entities=DottedDict(entities), interface=vass.speech)
            return Response(status=200)
        except KeyError:
            return Response(f"No such skill: {skill_name}", status=501)

    @app.route(f"/{name}/config", methods=["GET"])
    @authorized
    def get_config() -> Response:
        """Get Voice Assistant config."""
        return jsonify(Config)

    @app.route(f"/{name}/config", methods=["POST"])
    @authorized
    def set_config() -> Response:
        """Set Voice Assistant config.

        Payload must be a new config.
        """
        new_config = request.get_json() or {}
        try:
            Config.write(new_config)
            return Response(status=200)
        except ConfigValidationError:
            return Response("Invalid config", status=406)

    return app



"""Host Voice Assistant config respresentation."""

from typing import Dict

import yaml

from voiceassistant.const import CONFIG_FILE_PATH
from voiceassistant.utils.datastruct import DottedDict


class _Config(DottedDict):
    """Voice Assistant config parser."""

    def __init__(self, filepath: str) -> None:
        """Initialize config."""
        self.filepath = filepath
        self.reload()

    def reload(self) -> None:
        """Reload config."""
        super().__init__(self._read())

    def write(self, config_dict: Dict) -> None:
        """Write new config file from `config_dict`."""
        self._validate(config_dict)
        with open(self.filepath, "w") as config_file:
            yaml.dump(config_dict, config_file, default_flow_style=False)
        self.reload()

    def _read(self) -> Dict:
        """Get YAML config as dict."""
        with open(self.filepath) as config_file:
            config = yaml.load(config_file, Loader=yaml.FullLoader)
            return self._validate(config)

    def _validate(self, config: Dict) -> Dict:
        """Validate config."""
        return config


Config = _Config(CONFIG_FILE_PATH)



"""Voice assistant initial setup logic."""

import errno
import os

import yaml

from voiceassistant.const import (
    CONFIG_FILE_PATH,
    DEFAULT_CONFIG_DIR,
    GOOGLE_CREDENTIALS,
)


def pre_setup() -> None:
    """Do initial setup if needed."""
    if not os.path.exists(DEFAULT_CONFIG_DIR):
        try:
            os.makedirs(DEFAULT_CONFIG_DIR)
        except OSError as exc:
            if exc.errno != errno.EEXIST:
                raise

    sample_config = {
        "triggerword": {
            "picovoice": {"word": "jarvis", "sensitivity": 0.5},
            "sound": "now",
        },
        "prerecord_seconds": 3,
        "tts": {
            "aws": {
                "access_key_id": "YOUR_KEY_ID",
                "secret_access_key": "YOUR_SECRET_ACCESS_KEY",
                "region_name": "eu-west-2",
                "voice_id": "Brian",
            },
        },
        "stt": {"google_cloud": {"language_code": "en-US"}},
    }

    files_and_contents = (
        (GOOGLE_CREDENTIALS, ""),
        (CONFIG_FILE_PATH, yaml.dump(sample_config)),
    )

    for filepath, content in files_and_contents:
        if not os.path.exists(filepath):
            print(f"Creating {filepath}")
            with open(filepath, "w") as f:
                f.write(content)



"""Voice interface subpackage."""

from .keyword import KeywordDetector
from .microphone_stream import (
    MicrophoneStream,
    pause_microphone_stream,
    resume_microphone_stream,
)
from .speech_to_text import RecognitionString, SpeechToText
from .text_to_speech import TextToSpeech


class SpeechInterface:
    """Speech interface."""

    def __init__(self, rate: int):
        """Init."""
        self.sst = SpeechToText(rate)
        self.tts = TextToSpeech()

    def output(self, text: str, cache: bool = False) -> None:
        """Pronounce text."""
        self.tts.say(text, cache)

    def input(self) -> str:
        """Recognize speech."""
        pass


__all__ = [
    "KeywordDetector",
    "RecognitionString",
    "SpeechToText",
    "TextToSpeech",
    "SpeechInterface",
    "MicrophoneStream",
    "pause_microphone_stream",
    "resume_microphone_stream",
]



"""Add-On functions for speech interface."""

from usb.core import USBError

from voiceassistant.interfaces.speech import (
    pause_microphone_stream,
    resume_microphone_stream,
)

try:
    from pixel_ring import pixel_ring

    class PixelRingState:
        """Host pixel ring states."""

        off = 0
        speak = 1
        think = 2

    pixel_ring.off()
    ring_state = PixelRingState.off

    _mic_is_respeaker = True
except (USBError, ValueError, FileNotFoundError, PermissionError) as e:
    print(f"No ReSpeaker Microphone detected or not able to connect: {e}")
    _mic_is_respeaker = False


def processing_starts() -> None:
    """Do before NLP starts."""
    if _mic_is_respeaker:
        pixel_ring.speak()
        global ring_state
        ring_state = PixelRingState.speak


def processing_ends() -> None:
    """Do when NLP ends."""
    if _mic_is_respeaker:
        pixel_ring.off()
        global ring_state
        ring_state = PixelRingState.off


def tts_starts() -> None:
    """Do before voice output starts."""
    pause_microphone_stream()
    if _mic_is_respeaker:
        pixel_ring.think()


def tts_ends() -> None:
    """Do when voice output ends."""
    resume_microphone_stream()
    if _mic_is_respeaker:
        if ring_state == PixelRingState.speak:
            pixel_ring.speak()
        else:
            pixel_ring.off()



"""Natural language processor component."""

from types import TracebackType
from typing import Optional, Set, Type

from voiceassistant import addons
from voiceassistant.interfaces import InterfaceType
from voiceassistant.interfaces.speech import RecognitionString

from .nlp_result import NlpResult
from .regex import NLPregexProcessor

NLP_PROCESSORS = (NLPregexProcessor(),)


class NaturalLanguageProcessor:
    """Natural language processor class."""

    @addons.call_at(start=addons.speech.processing_starts)
    def __enter__(self):  # type: ignore
        """Start natural language processor."""
        self._processed_results: Set[NlpResult] = set()
        self._last_text_length = 0
        return self

    @addons.call_at(end=addons.speech.processing_ends)
    def __exit__(
        self,
        type: Type[BaseException],
        value: BaseException,
        traceback: TracebackType,
    ) -> Optional[bool]:
        """Stop natural language processor."""
        pass

    def process_single(self, transcript: str) -> None:
        """Process a single transcript.

        Use case: message from chatbot interface.
        """
        pass

    def _make_record(
        self, transcript: RecognitionString, nlp_result: NlpResult
    ) -> None:
        """Make record of a processed transcript."""
        self._processed_results.add(nlp_result)
        self._last_text_length = len(transcript)

    def _preprocess_transcript(
        self, text: RecognitionString
    ) -> RecognitionString:
        """Remove part of transcript that was already processed."""
        # fmt: off
        return RecognitionString(
            text[self._last_text_length:].lower(), is_final=text.is_final
        )
        # fmt: on

    def process_next_transcript(
        self, transcript: RecognitionString, interface: InterfaceType
    ) -> None:
        """Process a sequence of transcripts.

        Usecase: transcripts from continuos speech recognition.
        Each continuously recognized transcript is processed
        while user is speaking. Skill will be executed in two cases:
            1) transcript is complete (has enough information)
            2) transcript is final (user stopped speaking)
        """
        transcript = self._preprocess_transcript(transcript)

        for nlp_processor in NLP_PROCESSORS:
            nlp_result = nlp_processor.process(transcript)

            if not nlp_result:
                continue

            if nlp_result in self._processed_results:
                continue

            if nlp_result.is_complete or transcript.is_final:
                self._make_record(transcript, nlp_result)
                nlp_result.execute_skill(interface=interface)



"""Current time skill."""

from datetime import datetime

from voiceassistant.interfaces import InterfaceType
from voiceassistant.nlp.regex import regex_skill
from voiceassistant.utils.datastruct import DottedDict


@regex_skill(
    expressions=("(what&&time)",), entities={"location": ("Moscow", "London")},
)
def get_time(entities: DottedDict, interface: InterfaceType) -> None:
    """Get current time."""
    time_now = datetime.now().strftime("%H:%M")
    interface.output(f"It's {time_now}")


@regex_skill(expressions=("google for <<request>>",),)
def google_for(entities: DottedDict, interface: InterfaceType) -> None:
    """Google for somthing."""
    print(entities.request)



"""Home assistant control configurable skill."""

import hassapi

from voiceassistant.config import Config
from voiceassistant.interfaces import InterfaceType
from voiceassistant.nlp.regex import regex_skill
from voiceassistant.utils.datastruct import DottedDict
from voiceassistant.utils.hass import (
    get_entity_ids,
    hass_regex_skill,
    make_skill_func,
)

hass = hassapi.Hass(
    hassurl=Config.hass.url, token=Config.hass.token, timeout=20
)


@hass_regex_skill("turn_on", keywords=("on", "enable", "start"), hass=hass)
def turn_on(entities: DottedDict, interface: InterfaceType) -> None:
    """Turn on HASS entity."""
    for entity_id in get_entity_ids(entities.hass_entity_name):
        hass.turn_on(entity_id)
    interface.output(f"turning on the {entities.hass_entity_name}")


@hass_regex_skill("turn_off", keywords=("off", "disable"), hass=hass)
def turn_off(entities: DottedDict, interface: InterfaceType) -> None:
    """Turn off HASS entity."""
    for entity_id in get_entity_ids(entities.hass_entity_name):
        hass.turn_off(entity_id)
    interface.output(f"turning off the {entities.hass_entity_name}")


@hass_regex_skill("open_cover", keywords=("up", "open"), hass=hass)
def open_cover(entities: DottedDict, interface: InterfaceType) -> None:
    """Call open_cover service for HASS cover entity e.g. blinds."""
    for entity_id in get_entity_ids(entities.hass_entity_name):
        hass.open_cover(entity_id)
    interface.output(f"opening {entities.hass_entity_name}")


@hass_regex_skill("close_cover", keywords=("down", "close"), hass=hass)
def close_cover(entities: DottedDict, interface: InterfaceType) -> None:
    """Call close_cover service for HASS cover entity e.g. blinds."""
    for entity_id in get_entity_ids(entities.hass_entity_name):
        hass.close_cover(entity_id)
    interface.output(f"closing {entities.hass_entity_name}")


def _register_hass_custom_skills(hass: hassapi.Hass) -> None:
    """Register HASS custom skills specified in Config.hass.skills."""
    for skill in Config.hass.skills:
        # register skill
        regex_skill(expressions=skill.expressions, name=skill.name)(
            make_skill_func(skill.actions, hass)
        )


_register_hass_custom_skills(hass)



"""HTTP interface helper functions."""

import socket


def get_my_ip() -> str:
    """Get localhost IP address."""
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
        s.connect(("8.8.8.8", 80))
        return s.getsockname()[0]  # type: ignore



"""Host utilities for debugging."""

import sys

num_chars_printed = 0


def print_and_flush(text: str) -> None:
    """Print and flush."""
    global num_chars_printed
    overwrite_chars = " " * (num_chars_printed - len(text))
    sys.stdout.write(f"{text}{overwrite_chars}\r")
    sys.stdout.flush()
    num_chars_printed = len(text)



"""Utils for Home Assistant integration."""

from .actions import make_skill_func
from .entities import get_entity_ids, hass_regex_skill

__all__ = ["make_skill_func", "get_entity_ids", "hass_regex_skill"]



"""Host constant variables."""

import os

DEFAULT_CONFIG_DIR = f"{os.path.expanduser('~')}/.config/voiceassistant"
DATA_DIR = f"{os.path.dirname(os.path.realpath(__file__))}/data"

# must be set in order to construct google.cloud.speech.SpeechClient
if "GOOGLE_APPLICATION_CREDENTIALS" not in os.environ:
    os.environ[
        "GOOGLE_APPLICATION_CREDENTIALS"
    ] = f"{DEFAULT_CONFIG_DIR}/google_credentials.json"

GOOGLE_CREDENTIALS = os.environ["GOOGLE_APPLICATION_CREDENTIALS"]

CONFIG_FILE_PATH = os.environ.get(
    "VOICE_ASSISTANT_CONFIGURATION", f"{DEFAULT_CONFIG_DIR}/configuration.yaml"
)



"""Voice Assistant core components."""

import threading
import traceback

import voiceassistant.skills  # NOQA
from voiceassistant.config import Config
from voiceassistant.interfaces.http import HttpInterface
from voiceassistant.interfaces.speech import (
    KeywordDetector,
    MicrophoneStream,
    SpeechInterface,
)
from voiceassistant.nlp import NaturalLanguageProcessor
from voiceassistant.utils.debug import print_and_flush


class VoiceAssistant:
    """Voice Assistant root class."""

    def __init__(self) -> None:
        """Initialize Voice Assistant components."""
        self.keyword_detector = KeywordDetector()
        self.speech = SpeechInterface(rate=self.keyword_detector.rate)
        self.http = HttpInterface(self)

    def run(self) -> None:
        """Run Voice Assistant jobs in separate threads."""
        jobs = (
            self._speech_interface_loop,
            self.http.run,
        )
        for job in jobs:
            threading.Thread(target=job).start()

    def _speech_interface_loop(self) -> None:
        """Listen for keyword and process speech."""
        while True:
            print("Creating microphone stream")
            with MicrophoneStream(
                rate=self.keyword_detector.rate,
                chunk=self.keyword_detector.chunk_size,
                rolling_window_sec=Config.get("prerecord_seconds", 3),
            ) as stream:
                self.keyword_detector.wait_untill_detected(stream, self)

                with NaturalLanguageProcessor() as nlp:
                    try:
                        for (
                            transcript
                        ) in self.speech.sst.recognize_from_stream(stream):
                            print_and_flush(transcript)
                            nlp.process_next_transcript(
                                transcript=transcript, interface=self.speech
                            )
                    except Exception:
                        traceback.print_exc()
                        self.speech.output("Error occured", cache=True)



"""Utils for customizable HASS actions."""

import copy
from typing import Callable, List

import hassapi

from voiceassistant.interfaces import InterfaceType
from voiceassistant.utils.datastruct import DottedDict


def make_skill_func(actions: List[DottedDict], hass: hassapi.Hass) -> Callable:
    """Build skill function for a list of `actions` from Config."""

    def _hass_actions(entities: DottedDict, interface: InterfaceType) -> None:
        """Run actions in a loop."""
        for action_data in copy.deepcopy(actions):
            action_name = action_data.pop("action")
            print(f"running action {action_name} with data:\n{action_data}")
            func = _ACTIONS[action_name]
            func(hass, interface, action_data)

    return _hass_actions


def _say_from_template(
    hass: hassapi.Hass, interface: InterfaceType, data: DottedDict
) -> None:
    """Render HASS Jinja2 template and output result via `interface`."""
    interface.output(hass.render_template(template=data.template))


def _call_service(
    hass: hassapi.Hass, interface: InterfaceType, data: DottedDict
) -> None:
    """Call HASS service."""
    hass.call_service(
        service=data.pop("service"), entity_id=data.pop("entity_id"), **data
    )


def _fire_event(
    hass: hassapi.Hass, interface: InterfaceType, data: DottedDict
) -> None:
    """Fire HASS event."""
    hass.fire_event(event_type=data.event_type, event_data=data.event_data)


def _set_state(
    hass: hassapi.Hass, interface: InterfaceType, data: DottedDict
) -> None:
    """Set state of a HASS entity."""
    hass.set_state(entity_id=data.entity_id, state=data.state)


def _run_script(
    hass: hassapi.Hass, interface: InterfaceType, data: DottedDict
) -> None:
    """Run HASS script."""
    hass.run_script(data.script_id)


_ACTIONS = {
    "say_from_template": _say_from_template,
    "call_service": _call_service,
    "fire_event": _fire_event,
    "set_state": _set_state,
    "run_script": _run_script,
}



"""Keyword detector component."""

import struct

import pvporcupine

from voiceassistant.addons import keyword as keyword_addon
from voiceassistant.config import Config
from voiceassistant.exceptions import ConfigValidationError

from .microphone_stream import MicrophoneStream


class KeywordDetector:
    """Keyword detector."""

    def __init__(self) -> None:
        """Create Keyword detector object."""
        keyword = Config.triggerword.picovoice.word

        if keyword not in pvporcupine.KEYWORDS:
            raise ConfigValidationError(
                f"available keywords are {pvporcupine.KEYWORDS}"
            )

        self._detector = pvporcupine.create(
            keywords=[keyword],
            sensitivities=[Config.triggerword.picovoice.sensitivity],
        )

        self.rate = self._detector.sample_rate
        self.chunk_size = self._detector.frame_length

    def process(self, audio_chunk: bytes) -> int:
        """Process audio chunk.

        Returns:
            index of a recognized keyword if recognized, -1 otherwise
        """
        pcm = struct.unpack_from("h" * self.chunk_size, audio_chunk)
        return self._detector.process(pcm)  # type: ignore

    def not_detected(self, audio_chunk: bytes) -> bool:
        """Determine if keyword was not detected."""
        return self.process(audio_chunk) < 0

    def wait_untill_detected(  # type: ignore
        self, stream: MicrophoneStream, vass
    ) -> None:
        """Wait till keyword was detected."""
        while self.not_detected(stream.read()):
            pass
        keyword_addon.react_to_keyword(vass)



"""Utils for automatic creation of skills for HASS entities."""

from typing import Callable, Dict, Iterable, List, Set

import hassapi

from voiceassistant.config import Config
from voiceassistant.nlp.regex import regex_skill


def hass_regex_skill(
    service: str, keywords: Iterable[str], hass: hassapi.Hass
) -> Callable:
    """Wrap regex skill for convenience."""

    def wrapper(func: Callable) -> None:
        names = _get_names_with_service(service, hass)
        names_re = "|".join(names)
        keywords_re = "|".join(keywords)
        regex_skill(
            expressions=(f"({keywords_re}&&{names_re})",),
            entities={"hass_entity_name": names},
            name=f"hass_{service}",
        )(func)

    return wrapper


def get_entity_ids(name: str) -> List[str]:
    """Get HASS entity id list associated with a friendly name."""
    return _NAME_TO_IDS[name]


def _get_names_with_service(service: str, hass: hassapi.Hass) -> List[str]:
    """Get names of entities from Config for which ``service`` is applicable.

    Args:
        service: HASS service e.g. turn_on
        hass: hassapi client
    """
    domains = _get_domains_with_service(service, hass)
    names = (
        ent.names
        for ent in Config.hass.entities
        for id_ in ent.ids
        if id_.split(".")[0] in domains
    )
    return [name for sublist in names for name in sublist]  # flattened


def _get_domains_with_service(service: str, hass: hassapi.Hass) -> Set[str]:
    """Get HASS domains for which ``service`` is applicable."""
    return {
        svc.domain for svc in hass.get_services() if service in svc.services
    }


def _get_entity_name_to_ids() -> Dict[str, list]:
    """Get a mapping of name to list of HASS entity ids from Config."""
    result = {}
    for entity in Config.hass.entities:
        for name in entity.names:
            result[name] = entity.ids
    return result


_NAME_TO_IDS = _get_entity_name_to_ids()



"""Speech-to-text component."""

from typing import Generator

import google
from google.cloud import speech
from iterators import TimeoutIterator

from voiceassistant.config import Config
from voiceassistant.exceptions import SetupIncomplete
from voiceassistant.interfaces.speech.microphone_stream import MicrophoneStream


class RecognitionString(str):
    """Modify string object to have `is_final` attribute.

    Required to differentiate between cases where
    continious speech recognition is still running
    or whether recognition result is final.
    """

    is_final = False

    def __new__(cls, value: str, is_final: bool):  # type: ignore
        """Create recognition string object."""
        self_obj = str.__new__(cls, value)
        self_obj.is_final = is_final
        return self_obj


class SpeechToText:
    """Speech to Text class."""

    def __init__(self, rate: int):
        """Create speech-to-text object."""
        try:
            self._client = speech.SpeechClient()
        except google.auth.exceptions.DefaultCredentialsError as e:
            raise SetupIncomplete(e)

        config = speech.RecognitionConfig(
            encoding=speech.RecognitionConfig.AudioEncoding.LINEAR16,
            sample_rate_hertz=rate,
            language_code=Config.stt.google_cloud.language_code,
        )
        self._streaming_config = speech.StreamingRecognitionConfig(
            config=config, interim_results=True
        )

    def recognize_from_stream(
        self, stream: MicrophoneStream
    ) -> Generator[RecognitionString, None, None]:
        """Generate speech transcripts from audio stream."""
        is_final_once = False
        initial_transcript = ""
        transcript = "_empty_"

        requests = (
            speech.StreamingRecognizeRequest(audio_content=content)
            for content in stream.generator()
        )
        responses = self._client.streaming_recognize(
            self._streaming_config, requests, timeout=25  # type: ignore
        )
        responses = TimeoutIterator(responses, timeout=2)
        timedout_response = responses.get_sentinel()  # type: ignore

        for response in responses:
            # timeout in case user is not talking
            if response == timedout_response:
                print("*** STOPPING STT timeout***")
                yield RecognitionString(transcript, is_final=True)
                return

            if not response.results:
                continue

            result = response.results[0]
            if not result.alternatives:
                continue

            # best alternative
            transcript = (
                f"{initial_transcript} {result.alternatives[0].transcript}"
            )

            if not result.is_final:
                yield RecognitionString(transcript, is_final=False)
            else:
                if not is_final_once:
                    is_final_once = True
                    initial_transcript = transcript
                    yield RecognitionString(transcript, is_final=False)
                else:
                    print("*** STOPPING STT ***")
                    yield RecognitionString(transcript, is_final=True)
                    return



"""Host custom exceptions."""


class AssistantBaseException(Exception):
    """Base assistant exception."""


class ConfigValidationError(AssistantBaseException):
    """Config Validation Error."""


class NlpException(AssistantBaseException):
    """Natural Language Processor Error."""


class SetupIncomplete(AssistantBaseException):
    """Setup Incomplete Exception."""


class DottedAttribureError(AttributeError, AssistantBaseException):
    """Dotted Dictionary Attribute Error."""



"""Microphone audio stream implementation."""

from types import TracebackType
from typing import Dict, Generator, List, Tuple, Type

import pyaudio
from six.moves import queue

from voiceassistant.utils.datastruct import RollingWindowQueue

_PAUSED = False


def pause_microphone_stream() -> None:
    """Pause microphone stream."""
    global _PAUSED
    _PAUSED = True


def resume_microphone_stream() -> None:
    """Resume microphone stream."""
    global _PAUSED
    _PAUSED = False


class MicrophoneStream:
    """Opens a recording stream as a generator yielding the audio chunks."""

    def __init__(self, rate: int, chunk: int, rolling_window_sec: float = 3):
        """Create a microphone stream object."""
        self._rate = rate
        self._chunk = chunk

        # rolling window audio data buffer to store
        # the speech pronounced before the trigger word
        self._buff = RollingWindowQueue(
            size=int(rolling_window_sec * rate / chunk)
        )

        audio_interface = pyaudio.PyAudio()
        self._audio_stream = audio_interface.open(
            format=pyaudio.paInt16,
            channels=1,
            rate=self._rate,
            input=True,
            frames_per_buffer=self._chunk,
            stream_callback=self._fill_buffer,
        )
        self._last_chunks: List[bytes] = []
        self.closed = False

    def __enter__(self):  # type: ignore
        """Start audio stream."""
        return self

    def __exit__(
        self,
        type: Type[BaseException],
        value: BaseException,
        traceback: TracebackType,
    ) -> None:
        """Stop audio stream."""
        self._audio_stream.stop_stream()
        self._audio_stream.close()
        self.closed = True
        # Signal the generator to terminate so that the client's
        # streaming_recognize method will not block the process termination.
        self._buff.put(None)

    def _fill_buffer(
        self,
        in_data: bytes,
        frame_count: int,
        time_info: Dict,
        status_flags: int,
    ) -> Tuple:
        """Continuously collect data from the audio stream into the buffer."""
        if not _PAUSED:
            self._buff.put(in_data)
            self._last_chunks.append(in_data)
        return None, pyaudio.paContinue

    def read(self) -> bytes:
        """Get chunk of audio bytes."""
        while not self._last_chunks:
            pass
        chunk = b"".join(self._last_chunks)
        self._last_chunks = []
        return chunk

    def generator(self) -> Generator[bytes, None, None]:
        """Continuously generate chunks of audio data."""
        self._buff.disable_size_limit()

        while not self.closed:
            # Use a blocking get() to ensure there's at least one chunk of
            #  data, and stop iteration if the chunk is None, indicating the
            # end of the audio stream.
            chunk = self._buff.get()
            if chunk is None:
                return
            data = [chunk]

            # Now consume whatever other data's still buffered.
            while True:
                try:
                    chunk = self._buff.get(block=False)
                    if chunk is None:
                        return
                    data.append(chunk)
                except queue.Empty:
                    break

            yield b"".join(data)



"""Host logic for assistant specific regular expressions.

These expressions are regex but allow to specify entity names.
Example:
  - r"what's the weather in <location>"
  - r"google for <<query>>"

Single brackets <> specify single word entities.
Double brackets <<>> specify entities of few words.

And-operator (&&) is also added on top of regex
Example:
   - "(a&&b)" is equivalent to "(?=.*(?:a))(?=.*(?:b))"
"""

import re
from typing import Dict, Optional, Tuple, Union

from voiceassistant.exceptions import NlpException
from voiceassistant.utils.datastruct import DottedDict

_ENTITY_REGEX = re.compile(r"(<{1,2}.*?>{1,2})")
_LONG_ENTITY_REGEX = re.compile(r"(<<\w+>>)")
_SHORT_ENTITY_REGEX = re.compile(r"(<\w+>)")

_LONG_ENTITY_REPLACE = r"(.*)"
_SHORT_ENTITY_REPLACE = r"([a-zA-Z0-9_]*)"

_AND_OPERATOR = "&&"
_AND_OPERATOR_PATTERN = re.compile(r"\(.*&&.*\)")


class _FixedEntityName(str):
    """Entity of a fixed number of words."""

    fixed = True


class _VariableEntityName(str):
    """Entity of a variable number of words."""

    fixed = False


EntityNameType = Union[_FixedEntityName, _VariableEntityName]


class NlpExprMatch:
    """Nlp expression match class."""

    def __init__(self, entity_names: Tuple[EntityNameType, ...]):  # noqa
        """Init."""
        self.entity_names = entity_names
        self.entities = DottedDict()

    def __str__(self) -> str:
        """Get object string representation."""
        return f"{self.entities}"

    def update_entities(self, new: Dict) -> None:
        """Update entities dictionary."""
        self.entities.update(new)

    def set_last_entity_end(self, end: int) -> None:
        """Set the position of last entity end."""
        self.last_entity_end = end


class NLPregexExpression:
    """NLP expression parser."""

    def __init__(self, expression: str, entities: Optional[Dict]):
        """Create NLP regex expression."""
        self.expression = self._validate_expression(expression)
        self.regex = self._get_expression_regex(expression)
        self.entity_names = self._get_entity_names(expression)
        self.hard_entities = self._preprocess_hard_entities(entities)

    def _validate_expression(self, expression: str) -> str:
        """Validate assistant NLP expression."""
        return expression

    def _get_expression_regex(self, expression: str) -> re.Pattern:
        """Generate compiled regex from expression."""
        interpretations = (
            self._interpret_entity_names,
            self._interpret_and_operator,
        )
        for interpretation in interpretations:
            expression = interpretation(expression)
        return re.compile(expression)

    def _interpret_entity_names(self, expression: str) -> str:
        """Convert entity names specified in expression into regex form."""
        regex = re.sub(_LONG_ENTITY_REGEX, _LONG_ENTITY_REPLACE, expression)
        return re.sub(_SHORT_ENTITY_REGEX, _SHORT_ENTITY_REPLACE, regex)

    def _interpret_and_operator(self, expression: str) -> str:
        """Convert expression with && operator into regex form."""
        match = _AND_OPERATOR_PATTERN.search(expression)
        if match:
            parts = match.group(0)[1:-1].split(_AND_OPERATOR)
            return "".join(f"(?=.*(?:{part}))" for part in parts)
        return expression

    def _preprocess_hard_entities(
        self, entities: Optional[Dict]
    ) -> Optional[Dict]:
        """Convert keys of hardcoded `entities` dict into _FixedEntityName."""
        if entities:
            for value in entities.values():
                if not isinstance(value, (tuple, list)):
                    raise NlpException(
                        "Entities must be specified as tuples or lists"
                    )
            return {_FixedEntityName(k): v for k, v in entities.items()}
        return None

    def _get_entity_names(self, expression: str) -> Tuple[EntityNameType, ...]:
        """Extract entity names from NLP regex expression."""
        return tuple(
            self._parse_entity_name(name)
            for name in _ENTITY_REGEX.findall(expression)
        )

    def _parse_entity_name(self, entity_name: str) -> EntityNameType:
        """Parse and derive a type of an entity name."""
        if entity_name[0:2] == "<<":
            return _VariableEntityName(entity_name.strip("<>"))
        else:
            return _FixedEntityName(entity_name.strip("<>"))

    def _find_hard_entities(self, text: str) -> DottedDict:
        """Find hardcoded entities from a `text`."""
        result = DottedDict()

        for name, keywords in self.hard_entities.items():  # type: ignore
            found = [ent for ent in keywords if ent in text]
            if found:
                result.update({name: found[0]})
        return result

    def match(self, text: str) -> Optional[NlpExprMatch]:
        """Match and extract entities from `text`."""
        matched = self.regex.search(text)

        if matched:
            nlp_expr_match = NlpExprMatch(entity_names=self.entity_names)
            entities = matched.groups()

            if entities:
                if len(entities) != len(self.entity_names):
                    raise NlpException("Matched unexpected number of entities")

                name_to_entity = dict(zip(self.entity_names, entities))
                nlp_expr_match.update_entities(name_to_entity)
                nlp_expr_match.set_last_entity_end(
                    matched.end(matched.lastindex)  # type: ignore
                )

            if self.hard_entities:
                nlp_expr_match.update_entities(self._find_hard_entities(text))

            return nlp_expr_match
        return None



"""Main file."""

from voiceassistant.exceptions import SetupIncomplete
from voiceassistant.setup import pre_setup


def main() -> None:
    """Run application."""
    pre_setup()

    from voiceassistant.core import VoiceAssistant

    try:
        vass = VoiceAssistant()
        vass.run()
    except SetupIncomplete as e:
        print(f"Voice Assistant setup is incomplete:\n{e}")


if __name__ == "__main__":
    main()



"""Host regex based NL processor and skill wrapper."""

from typing import Optional

from ..nlp_result import NlpResult
from .skill import REGEX_SKILLS, regex_skill


class NLPregexProcessor:
    """NL regex processor."""

    def process(self, transcript: str) -> Optional[NlpResult]:
        """Process transcript by matching it to each skill."""
        for skill in REGEX_SKILLS.values():
            nlp_result = skill.match(transcript)
            if nlp_result:
                return nlp_result
        return None


__all__ = ["regex_skill", "NLPregexProcessor"]



"""Host regex skill struct."""
import warnings
from typing import Callable, Dict, Iterable, Optional, Set

from ..nlp_result import NlpResult
from .expression import NlpExprMatch, NLPregexExpression

REGEX_SKILLS = {}


def regex_skill(
    expressions: Iterable[str],
    entities: Optional[Dict] = None,
    name: Optional[str] = None,
) -> Callable:
    """Regex based skill wrapper."""

    def wrapper(func: Callable) -> Callable:
        skill_name = name or func.__name__
        if skill_name in REGEX_SKILLS:
            warnings.warn(
                f"Skill {skill_name} is already registered, skipping"
            )
        else:
            REGEX_SKILLS[skill_name] = RegexSkillStruct(
                name=skill_name,
                func=func,
                expressions=expressions,
                entities=entities,
            )
            print(f"Skill added: {skill_name}")
        return func

    return wrapper


class RegexSkillStruct:
    """Store regex skill related attributes."""

    def __init__(
        self,
        name: str,
        func: Callable,
        expressions: Iterable[str],
        entities: Optional[Dict] = None,
    ) -> None:
        """Create regex skill struct."""
        self.name = name
        self.func = func
        self.expressions = tuple(
            NLPregexExpression(expr, entities) for expr in expressions
        )

    def __str__(self) -> str:
        """Get string respresentation of self."""
        return f"Skill: {self.func.__name__}"

    def _get_all_entity_names(self) -> Set:
        """Get all skill related entity names."""
        return {ent for expr in self.expressions for ent in expr.entity_names}

    def _is_complete(self, text: str, match: NlpExprMatch) -> bool:
        """Determine if text has enough info to be processed."""
        # regex skill has no entities at all
        if not match.entity_names:
            return True
        # last matched entity is fixed
        if match.entity_names[-1].fixed:
            return True
        # all found entities are fixed
        if all(ent.fixed for ent in match.entity_names):
            return True
        # if string is longer than postion of last entity
        if match.last_entity_end < len(text):
            return True
        # all potential entities are present?
        # if self.all_entity_names == set(match.entities):
        #     return True
        return False

    def match(self, text: str) -> Optional[NlpResult]:
        """Match text to this skill."""
        for expr in self.expressions:
            expr_match = expr.match(text)
            if expr_match:
                return NlpResult(
                    skill_name=self.name,
                    skill_func=self.func,
                    entities=expr_match.entities,
                    is_complete=self._is_complete(text, expr_match),
                )
        return None



"""Host custom data structures used in the project."""

from typing import Any

from six.moves import queue

from voiceassistant.exceptions import DottedAttribureError


class DottedDict(dict):
    """Python dict with dot-syntax added on top."""

    def __getattr__(self, key: str) -> Any:
        """Get attribute."""
        try:
            attr = self[key]
            if isinstance(attr, dict):
                return DottedDict(attr)
            elif isinstance(attr, list):
                return [
                    DottedDict(i) if isinstance(i, dict) else i for i in attr
                ]
            else:
                return attr
        except KeyError as e:
            raise DottedAttribureError(e)

    def __setattr__(self, key: str, value: Any) -> None:
        """Set attribute."""
        try:
            self[key] = value
        except KeyError as e:
            raise DottedAttribureError(e)


class RollingWindowQueue:
    """Rolling window queue."""

    def __init__(self, size: int):
        """Init."""
        self.size = size
        self._current_size = 0
        self._buffer: queue.Queue = queue.Queue()
        self._size_limit_enabled = True

    def disable_size_limit(self) -> None:
        """Convert to unlimited size Queue."""
        self._size_limit_enabled = False

    def put(self, element: Any) -> None:
        """Put element to the queue."""
        self._buffer.put(element)
        self._current_size += 1
        if self._size_limit_enabled and self._current_size > self.size:
            self._buffer.get()

    def get(self, block: bool = True) -> Any:
        """Get element from the queue."""
        self._current_size -= 1
        return self._buffer.get(block=block)



"""HTTP interface subpackage."""

from typing import Callable

from flask import Flask

from .api_app import api_factory
from .helpers import get_my_ip

PORT = 5050


class HttpInterface:
    """HTTP interface."""

    def __init__(self, vass) -> None:  # type: ignore
        """Initialize flask app."""
        self.vass = vass

        self.app = Flask(__name__)
        self.register_factories(api_factory)

    def register_factories(self, *factories: Callable) -> None:
        """Register Flask app factories."""
        for factory in factories:
            self.app = factory(vass=self.vass, app=self.app)

    def run(self) -> None:
        """Run app."""
        self.app.run(get_my_ip(), port=PORT)



"""Handle module packaging."""

import subprocess
import textwrap

from setuptools import setup


def _get_release_version() -> None:
    """Get git release tag version."""
    version = (
        subprocess.run(["git", "describe", "--tags"], stdout=subprocess.PIPE)
        .stdout.decode("utf-8")
        .strip()
    )
    print("VERSION: ", version)
    # assert re.match(r"\d\.\d.\d{1,2}", version)
    return version


def _write_version(version: str) -> None:
    """Write version to version.py."""
    with open("src/voiceassistant/version.py", "w") as file:
        file.write(
            textwrap.dedent(
                f'''
                """Host package version, generated on build."""
                __version__ = "{version}"
                '''
            ).lstrip()
        )


version = _get_release_version()

_write_version(version)
setup(version=version)



"""Text-to-speech component."""

import os

import boto3
from botocore.config import Config as BotoConfig
from botocore.exceptions import ClientError, NoCredentialsError

from voiceassistant import addons
from voiceassistant.addons import speech as speech_addon
from voiceassistant.config import Config
from voiceassistant.const import DEFAULT_CONFIG_DIR
from voiceassistant.exceptions import SetupIncomplete


class TextToSpeech:
    """Text to Speech class using Amazon Polly."""

    def __init__(self) -> None:
        """Create text-to-speech object."""
        boto_config = BotoConfig(
            region_name=Config.tts.aws.region_name,
            connect_timeout=0.7,
            read_timeout=0.7,
            parameter_validation=False,
        )
        self._client = boto3.Session(
            aws_access_key_id=Config.tts.aws.access_key_id,
            aws_secret_access_key=Config.tts.aws.secret_access_key,
        ).client(service_name="polly", config=boto_config)

        try:
            self._client.describe_voices()
        except (NoCredentialsError, ClientError):
            raise SetupIncomplete("Amazon Polly credentials not set")

    @addons.call_at(start=speech_addon.tts_starts, end=speech_addon.tts_ends)
    def say(self, text: str, cache: bool = False) -> None:
        """Pronounce `text` with configured Polly voice.

        ToDo:
            Current implementation is bad and should later
            be changed to play audio bytes directly from memory.
        """
        if cache:
            filename = text.replace(" ", "")

            if not os.path.isfile(f"{DEFAULT_CONFIG_DIR}/{filename}.mp3"):
                self.synthesize_to_mp3_file(text, filename=filename)
                print(f"TTS cached: {text}")

            self._play_mp3_file(filename)
        else:
            self.synthesize_to_mp3_file(text, filename="speech")
            self._play_mp3_file("speech")
            os.system(f"rm {DEFAULT_CONFIG_DIR}/speech.mp3")

    def _play_mp3_file(self, filename: str) -> None:
        """Play mp3 file."""
        os.system(
            f"mpg123 {DEFAULT_CONFIG_DIR}/{filename}.mp3 >/dev/null 2>&1"
        )

    def synthesize_to_mp3_file(self, text: str, filename: str) -> None:
        """Save synthesized `text` to mp3 file."""
        audio_bytes = self.synthesize(text, format="mp3")
        with open(f"{DEFAULT_CONFIG_DIR}/{filename}.mp3", "wb") as file:
            file.write(audio_bytes)

    def synthesize(self, text: str, format: str = "mp3") -> bytes:
        """Synthesize `text` to audio bytes."""
        return self._client.synthesize_speech(  # type: ignore
            VoiceId=Config.tts.aws.voice_id, OutputFormat=format, Text=text
        )["AudioStream"].read()
        # ogg_vorbis



"""Handle HTTP server authentication."""

import secrets
import string
from functools import wraps
from typing import Any, Callable

import flask
from flask import request

from voiceassistant.const import DEFAULT_CONFIG_DIR


def _get_token() -> str:
    """Create if needed and return API token."""
    tokenfile = f"{DEFAULT_CONFIG_DIR}/API_TOKEN"

    try:
        with open(tokenfile) as file:
            return file.read()
    except FileNotFoundError:
        token = _generate_token()

        print(f"\n\nPlease save auto-generated API token: {token}\n\n")

        with open(tokenfile, "w") as file:
            file.write(token)

        return token


def _generate_token(n_chars: int = 50) -> str:
    """Generate API token."""
    alphabet = string.ascii_letters + string.digits
    return "".join(secrets.choice(alphabet) for i in range(n_chars))


TOKEN = _get_token()


def authorized(func: Callable) -> Callable:
    """Wrap Flask handler function to check for auth TOKEN."""

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        if TOKEN == request.headers.get("token"):
            return func(*args, **kwargs)
        else:
            print("Unauthorized")
            return flask.Response("<b>401: Unauthorized</b>", status=401)

    return wrapper
